#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define HORI    0
#define VERT    1
#define BLOCK   2

struct swordfish_struct { int candidate; 
                          int x[ 9 ];
                          int y[ 9 ];
                        } ;

 
extern int i_puzzle[9][9][10];
extern int debug;

// for statistics
int  used_swordfish[2] = { 0, 0 }; 
int  used_x_wing[2] = { 0, 0 }; 
extern void dump_rc( int, int, int );
extern void show_candidate ( int ) ;
extern void check_possibilities_s( void );  // singles
extern int  elim_possibility( int, int, int );

void x_wing ( void );
void swordfish ( void );
int  swordfish_rc ( struct swordfish_struct [], int [], int , int );




/****************************************************************************
 * x_wing(): detect x_wing patterns                                         *
 ****************************************************************************/
void x_wing ( void )
{
    int x, x_s, y, y_s, z, i, j, i_xw, candidates_found, candidate, found[2], dir, xory, xory_s; 
    struct x_wing_struct { int candidate; 
                           int x[ 9 ];
                           int y[ 9 ];
                          } s_xwing[ 9 ] ;

    
    for ( dir = HORI; dir <= VERT; dir++ ) {    
        for ( candidate = 1; candidate <= 9; candidate++ ) {
            i_xw = 0;
            x    = 0;
            y    = 0;

            for ( i = 0; i < 9; i++ ) {
                for ( j = 0; j < 9; j++ ) {
                    s_xwing[i].x[j] = s_xwing[i].y[j] = -1;
                }
            }
            
            while (( x < 9 ) && ( y < 9 )) {
                candidates_found = 0;
                while (( x < 9 ) && ( y < 9 )) {
                    if ( i_puzzle[x][y][0] == 0 ) { // no need to look in determined cells
                        for ( z = 1; z < 10; z++ ) {
                            if ( i_puzzle[x][y][z] == candidate ) {
                                s_xwing[ i_xw ].candidate = candidate;
                                s_xwing[ i_xw ].x[ candidates_found ] = x;
                                s_xwing[ i_xw ].y[ candidates_found ] = y;
                                candidates_found++;
                            }
                        }
                    }
                    if ( dir == VERT ) {
                        x++;
                    }
                    else if ( dir == HORI ) {
                        y++;
                    }
                }
                if ( candidates_found == 2 ) {
                    i_xw++ ;
                }
                if ( dir == VERT ) {
                    y++;
                    x = 0;
                }
                else if ( dir == HORI ) {
                    x++;
                    y = 0;
                }
            } 
    
            if ( debug > 3 ) {                    
                for ( i = 0; i < i_xw; i++ ) {
                    printf ( "x_wing(): s_xw[%d].can=%d i_xw=%d dir=%s ", i, s_xwing[i].candidate, i_xw, which_dir( dir ) ); 
                    for ( j = 0; j < 5; j++ ) {
                        if ( dir == VERT ) {
                            if ( s_xwing[i].x[j] != -1 ) { 
                                printf( "s_xw[%d].x[%d]=%2d ", i, j, s_xwing[i].x[j] );
                            }
                        }
                        else if ( dir == HORI ) {
                            if ( s_xwing[i].y[j] != -1 ) { 
                               printf( "s_xw[%d].y[%d]=%2d ", i, j, s_xwing[i].y[j] );
                            }
                        }
                    }
                    putchar ('\n');
                }
            }
    
            if ( i_xw >= 2 ) {  // we've possibly found an X-wing
                if ( dir == VERT ) {
                    found[ 0 ] = -1;
                    for ( y = 0; y < (i_xw - 1); y++ ) {
                        for ( y_s = ( y + 1 ); y_s < i_xw; y_s++ ) {
                            if (( s_xwing[ y ].x[0] == s_xwing[ y_s ].x[0] ) && ( s_xwing[ y ].x[1] == s_xwing[ y_s ].x[1] )) {
                                if ( found[ 0 ] == -1 ) {
                                    found[ 0 ] = y;
                                    found[ 1 ] = y_s;
                                }
                                else {  // we've got too many hits!
                                    found[ 0 ] = -1;
                                    break;
                                }
                            }
                        }
                    }
                }
                else if ( dir == HORI ) {
                    found[ 0 ] = -1;
                    for ( x = 0; x < (i_xw - 1); x++ ) {
                        for ( x_s = ( x + 1 ); x_s < i_xw; x_s++ ) {
                            if (( s_xwing[ x ].y[0] == s_xwing[ x_s ].y[0] ) && ( s_xwing[ x ].y[1] == s_xwing[ x_s ].y[1] )) {
                                if ( found[ 0 ] == -1 ) {
                                    found[ 0 ] = x;
                                    found[ 1 ] = x_s;
                                }
                                else {  // we've got too many hits!
                                    found[ 0 ] = -1;
                                    break;
                                }
                            }
                        }
                    }
                }
            
                if ( found[ 0 ] >= 0 )  {
                    if ( debug > 1 ) {
                        printf ( "Found an X-wing in (%d,%d), (%d,%d), (%d,%d), (%d,%d) can=%d dir=%s\n",\
                                                      s_xwing[ found[0] ].x[0], s_xwing[ found[0] ].y[0],\
                                                      s_xwing[ found[0] ].x[1], s_xwing[ found[0] ].y[1],\
                                                      s_xwing[ found[1] ].x[0], s_xwing[ found[1] ].y[0],\
                                                      s_xwing[ found[1] ].x[1], s_xwing[ found[1] ].y[1],\
                                                      s_xwing[0].candidate, which_dir ( dir ));  
                        if ( debug > 2 ) {
                            show_candidate ( s_xwing[0].candidate );
                        }
                    }

                    if ( dir == VERT ) {
                        for ( y = 0; y < 9; y ++ ) {
                            if (( y != s_xwing[found[0]].y[0] ) && ( y != s_xwing[found[1]].y[0] )) {
                                if ( debug > 3 ) {
                                    puts( "x_wing(): calling elim_possibility()");
                                }
                                used_x_wing[ dir ] += elim_possibility ( s_xwing[found[0]].x[0], y, candidate );
                                used_x_wing[ dir ] += elim_possibility ( s_xwing[found[0]].x[1], y, candidate );
                            }
                        }
                    }
                    else if ( dir == HORI ) {
                        for ( x = 0; x < 9; x ++ ) {
                            if (( x != s_xwing[found[0]].x[0] ) && ( x != s_xwing[found[1]].x[0] )) {
                                if ( debug > 3 ) {
                                    puts( "x_wing(): calling elim_possibility()");
                                }
                                used_x_wing[ dir ] += elim_possibility ( x, s_xwing[found[0]].y[0], candidate );
                                used_x_wing[ dir ] += elim_possibility ( x, s_xwing[found[0]].y[1], candidate );
                            }
                        }
                    }
                }
            }  // if i_xw
            
        }  // for candidate 
            
    } // for dir
}
 


/******************************************************************
 * swordfish detection                                            *
 ******************************************************************/
void swordfish ( void )
{
    int x, y, xy_s[5], z, i, j, i_sf, candidates_found, candidate, found[2], dir, xory, xory_s; 
    int extension; // for swordfishes with more than 3 cols/rows
    struct swordfish_struct s_sf [ 9 ] ;

    
    for ( dir = HORI; dir <= VERT; dir++ ) {    
        for ( extension = 3; extension <= 5; extension++ ) {
            /////// debugging
     //       extension = 3;
     //       dir = VERT;
            ///////// end debugging
            for ( candidate = 1; candidate <= 9; candidate++ ) {
                i_sf = 0;
                x    = 0;
                y    = 0;
                for ( i = 0; i < 9; i++ ) {
                    for ( j = 0; j < 9; j++ ) {
                        s_sf[i].x[j] = s_sf[i].y[j] = -1;
                    }
                }

/* for debugging
                if ( candidate == 2 ) {
                    dump_rc ( 0, 0, HORI );
                    show_candidate (2);
                    i_puzzle[7][0][0]=0;
                    i_puzzle[7][0][1]=1;
                    i_puzzle[7][0][2]=2;
                    i_puzzle[1][0][0]=0;
                    i_puzzle[1][0][1]=1;
                    i_puzzle[1][0][2]=2;
                    dump_rc ( 0, 0, HORI );
                    show_candidate (2);
                }
end for debugging */
                
                while (( x < 9 ) && ( y < 9 )) {
                    candidates_found = 0;
                    while (( x < 9 ) && ( y < 9 )) {
                        if ( i_puzzle[x][y][0] == 0 ) { // no need to look in determined cells
                            for ( z = 1; z < 10; z++ ) {
                                if ( i_puzzle[x][y][z] == candidate ) {
                                    s_sf[ i_sf ].candidate = candidate;
                                    s_sf[ i_sf ].x[ candidates_found ] = x;
                                    s_sf[ i_sf ].y[ candidates_found ] = y;
                                    candidates_found++;
                                }
                            }
                        }
                        if ( dir == VERT ) {
                            x++;
                        }
                        else if ( dir == HORI ) {
                            y++;
                        }
                    }
                    if (( candidates_found > 0 ) && ( candidates_found <= extension )) {
                        i_sf++ ;
                    }
                    if ( dir == VERT ) {
                        y++;
                        x = 0;
                    }
                    else if ( dir == HORI ) {
                        x++;
                        y = 0;
                    }
                } 
        
                if ( debug > 3 ) {
                    for ( i = 0; i < i_sf; i++ ) {
                        printf ( "swordfish(): s_sf[%d].can=%d i_sf=%d dir=%s ext=%d ", i, s_sf[i].candidate, i_sf, which_dir( dir ), extension ); 
                        for ( j = 0; j < 5; j++ ) {
                            if ( dir == VERT ) {
                                if ( s_sf[i].x[j] != -1 ) { 
                                    printf( "s_sf[%d].x[%d]=%2d ", i, j, s_sf[i].x[j] );
                                }
                            }
                            else if ( dir == HORI ) {
                                if ( s_sf[i].y[j] != -1 ) { 
                                    printf( "s_sf[%d].y[%d]=%2d ", i, j, s_sf[i].y[j] );
                                }
                            }
                        }
                        putchar ('\n');
                    }
                    if ( dir == VERT ) {
                        if ( candidate == 1) {
                            show_candidate(1);
                        }
                    }
                }
        
                if ( i_sf >= extension ) {  // we've possibly found a swordfish
                    for ( xy_s[0] = 0; xy_s[0] < ( ( i_sf + 1 ) - extension ); xy_s[0]++ ) {
                        for ( xy_s[1] = ( xy_s[0] + 1 ); xy_s[1] < (( i_sf + 2 ) - extension ); xy_s[1]++ ) {
                            for ( xy_s[2] = ( xy_s[1] + 1 ); xy_s[2] < (( i_sf + 3 ) - extension ); xy_s[2]++ ) {
                                if ( extension > 3 ) {
                                    for ( xy_s[3] = ( xy_s[2] + 1 ); xy_s[3] < (( i_sf + 4 ) - extension ); xy_s[3]++ ) {
                                        if ( extension > 4 ) {
                                            for ( xy_s[4] = ( xy_s[3] + 1 ); xy_s[4] < (( i_sf + 5 ) - extension ); xy_s[4]++ ) {
                                                if ( swordfish_rc ( s_sf, xy_s, 5, dir ) > 0 ) { 
                                                    if ( debug > 1 ) {
                                                        printf("swordfish(): Found!\n");
                                                        show_candidate( candidate );
                                                    }   
                                                }
                                            }
                                        }
                                        else {
                                            if ( swordfish_rc ( s_sf, xy_s, 4, dir ) > 0 ) { 
                                                if ( debug > 1 ) {
                                                    printf("swordfish(): Found!\n");
                                                    show_candidate( candidate );
                                                }   
                                            }
                                        }
                                    }
                                }
                                else {        
                                    if ( swordfish_rc ( s_sf, xy_s, 3, dir ) > 0 ) { 
                                        if ( debug > 1 ) {
                                            printf("swordfish(): Found!\n");
                                            show_candidate( candidate );
                                        }   
                                    }
                                }
                            }
                        }
                    }
                
                    if ( 1 == 0 )  {
                        if ( debug > 1 ) {
                            printf ( "Found a Swordfish in (%d,%d), (%d,%d), (%d,%d), (%d,%d) can=%d dir=%s\n",\
                                                          s_sf[ found[0] ].x[0], s_sf[ found[0] ].y[0],\
                                                          s_sf[ found[0] ].x[1], s_sf[ found[0] ].y[1],\
                                                          s_sf[ found[1] ].x[0], s_sf[ found[1] ].y[0],\
                                                          s_sf[ found[1] ].x[1], s_sf[ found[1] ].y[1],\
                                                          s_sf[0].candidate, which_dir ( dir ));  
                            if ( debug > 2 ) {
                                show_candidate ( s_sf[0].candidate );
                            }
                        }
    
                        if ( dir == VERT ) {
                            for ( y = 0; y < 9; y ++ ) {
                                if (( y != s_sf[found[0]].y[0] ) && ( y != s_sf[found[1]].y[0] )) {
                                    if ( debug > 3 ) {
                                        puts( "swordfish(): calling elim_possibility()");
                                    }
                                    used_swordfish[ dir ] += elim_possibility ( s_sf[found[0]].x[0], y, candidate );
                                    used_swordfish[ dir ] += elim_possibility ( s_sf[found[0]].x[1], y, candidate );
                                }
                            }
                        }
                        else if ( dir == HORI ) {
                            for ( x = 0; x < 9; x ++ ) {
                                if (( x != s_sf[found[0]].x[0] ) && ( x != s_sf[found[1]].x[0] )) {
                                    if ( debug > 3 ) {
                                        puts( "swordfish(): calling elim_possibility()");
                                    }
                                    used_swordfish[ dir ] += elim_possibility ( x, s_sf[found[0]].y[0], candidate );
                                    used_swordfish[ dir ] += elim_possibility ( x, s_sf[found[0]].y[1], candidate );
                                }
                            }
                        }
                    }
                }  // if i_sf
            }  // for candidate 
        } // for extension
    } // for dir
}


/************************************************************************
 * swordfish_rc(): find out in which columns we have our possibilities  *
 *                                                                      *
 * return 0: no swordfish                                               *
 *       >0: number of cols used                                        *
 *                                                                      *
 ************************************************************************/
int swordfish_rc( struct swordfish_struct s_sf[], int xy_s[], int extension, int dir )
{
    int rowcols_used[9] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
    int ret = 0;
    int i, j;
    int rowcols_total = 0;
   
    for ( i = 0; i < extension; i++ ) {
        j = 0;
        if ( dir == VERT ) {
            while ( s_sf[ xy_s[i] ].x[ j ] != -1 ) {
                rowcols_used[ s_sf[ xy_s[i] ].x[ j ] ]++;
                j++;
            }
        }
        else if ( dir == HORI ) {
            while ( s_sf[ xy_s[i] ].y[ j ] != -1 ) {
                rowcols_used[ s_sf[ xy_s[i] ].y[ j ] ]++;
                j++;
            }
        }
        else {
            puts( "swordfish_rc(): unknown direction! (bug)" );
        }
    }
    
    if ( debug > 3 ) {
        for ( i = 0; i < extension; i++ ) {
            printf ( "xy_s[%d]=%d ", i, xy_s[i] );
        }
        putchar ('\n');
        if ( dir == VERT ) {
            printf( "swordfish_rc(): can=%d ext=%d cols used: ", s_sf[0].candidate, extension );
        }
        else if ( dir == HORI ) {
            printf( "swordfish_rc(): can=%d ext=%d rows used: ", s_sf[0].candidate, extension );
        }
    }
    for ( i = 0; i < 9; i++ ) {
        if ( rowcols_used[i] > 0 ) {
            rowcols_total++;
            if ( debug > 0 ) {
                printf( "%d:%d ", i, rowcols_used[i] );
            } 
        }
    }
    if ( debug > 0 ) {
        putchar( '\n' );
    }
    
    if ( rowcols_total <= extension ) {
        ret = rowcols_total;
    }

    return( ret );

}

